from enum import Enum

class DataFormat(Enum):
    """
    The possible data format types for record field values.
    """

    Raw = 0
    Formatted = 1

class ReportDataType(Enum):
    """
    The possible report data types for reports.
    """
    ReportData = 0
    ChartData = 1

class ResultValueType(Enum):
    """
    The possible types for record field values.
    """
    String = 0
    Integer = 1
    Decimal = 2
    Date = 3
    TimeSpan = 4
    Guid = 5
    StringList = 6
    IntegerList = 7
    GuidList = 8
    AttachmentList = 9
    ScoringGroupList = 10
    FileList = 11

class Increment(Enum):
    """
    The possible values for the increment property of timespan data in an Onspring timespan field.
    """
    Seconds = "Second(s)"
    Minutes = "Minute(s)"
    Hours = "Hour(s)"
    Days = "Day(s)"
    Weeks = "Week(s)"
    Months = "Month(s)"
    Years = "Year(s)"

class Recurrence(Enum):
    """
    The possible values for the recurrence property of timespan data in an Onspring timespan field.
    """
    Empty = "None"
    EndByDate = "EndByDate"
    EndAfterOccurrences = 'EndAfterOccurrences'




import requests
import json
import re

from OnspringApiSdk.Endpoints import *
from OnspringApiSdk.Models import *

class OnspringClient:
    """
    A class that represents a client that can interact with the api.
    
    Attributes:
        baseUrl (`str`): The url that should be used as the base for all requests made by the client.
        headers (`dict`): Contains key value pairs of the headers necessary for all requests made by the client including the necessary api key.
    """
    def __init__(self, url: str, key: str):
        self.baseUrl = url
        self.headers = {
            'x-apikey': key,
            'x-api-version': '2'
        }

    # connectivity methods

    def CanConnect(self):
        """
        Verifies if the API is reachable by calling the ping endpoint.

        Args:
            None

        Returns:
            A `bool` value indicating if the API is responsive.
        """
        endpoint = GetPingEndpoint(self.baseUrl)

        response = requests.request(
            'GET',
            endpoint,
            headers=self.headers)

        return response.status_code == 200

    # app methods

    def GetApps(self, pagingRequest=PagingRequest(1, 50)):
        """
        Gets all accessible apps for client.

        Parameters:
            pagingRequest (`Models.PagingRequest`): Used to set the page number and page size of the request. By default the these will be 1 and 50 respectively.

        Returns:
            An ApiResponse (`Models.ApiResponse`) containing the results of the request.
        """

        endpoint = GetAppsEndpoint(self.baseUrl)

        params = pagingRequest.__dict__

        response = requests.request(
            'GET',
            endpoint,
            headers=self.headers,
            params=params)

        if response.status_code == 400:
            return ApiResponse(
                response.status_code,
                message='Invalid paging information',
                raw=response)

        if response.status_code == 401:
            return ApiResponse(
                response.status_code,
                message='Unauthorized request',
                raw=response)

        if response.status_code == 200:

            responseJson = dict(response.json())

            apps = []

            for item in responseJson.get('items'):

                item = dict(item)

                app = App(
                    item.get('href'),
                    item.get('id'),
                    item.get('name'))

                apps.append(app)

            data = GetAppsResponse(
                responseJson.get('pageNumber'),
                responseJson.get('pageSize'),
                responseJson.get('totalPages'),
                responseJson.get('totalRecords'),
                apps)

            return ApiResponse(
                response.status_code,
                data,
                raw=response)

        return ApiResponse(
            response.status_code,
            raw=response)

    def GetAppById(self, appId: int):
        """
        Get an app by it's id.

        Parameters:
            appId (`int`): The unique id of the app being requested.

        Returns:
            An ApiResponse (`Models.ApiResponse`) containing the results of the request.
        """

        endpoint = GetAppByIdEndpoint(self.baseUrl, appId)

        response = requests.request(
            'GET',
            endpoint,
            headers=self.headers)

        if response.status_code == 401:
            
            return ApiResponse(
                response.status_code,
                message='Unauthorized request',
                raw=response)

        if response.status_code == 403:
            
            return ApiResponse(
                response.status_code,
                message='Client does not have read access to the app',
                raw=response)

        if response.status_code == 404:
            
            return ApiResponse(
                response.status_code,
                message='App could not be found',
                raw=response)

        if response.status_code == 200:

            responseJson = dict(response.json())

            app = App(
                responseJson.get('href'),
                responseJson.get('id'),
                responseJson.get('name'))

            data = GetAppByIdResponse(app)

            return ApiResponse(
                response.status_code,
                data,
                raw=response)

        return ApiResponse(
            response.status_code,
            raw=response)

    def GetAppsByIds(self, appIds: list):
        """
        Get a set of apps by their ids.

        Parameters:
            appIds (`list[int]`): A list of unique ids for the apps being requested.

        Returns:
            An ApiResponse (`Models.ApiResponse`) containing the results of the request.
        """

        endpoint = GetAppsByIdsEndpoint(self.baseUrl)

        self.headers['Content-Type'] = 'application/json'

        # make sure appIds can be serialized to json string
        if not isinstance(appIds, (list, tuple)):
            return ApiResponse(
                400,
                message='App ids should be of type list or tuple')

        appIds = json.dumps(appIds)

        response = requests.request(
            'POST',
            endpoint,
            headers=self.headers,
            data=appIds)

        if response.status_code == 401:
            
            return ApiResponse(
                response.status_code,
                message='Unauthorized request',
                raw=response)

        if response.status_code == 403:
            
            return ApiResponse(
                response.status_code,
                message='Client does not have read access to the app',
                raw=response)

        if response.status_code == 200:

            responseJson = dict(response.json())

            apps = []

            for item in responseJson['items']:

                item = dict(item)

                app = App(
                    item.get('href'),
                    item.get('id'),
                    item.get('name'))

                apps.append(app)

            data = GetAppsByIdsResponse(
                responseJson.get('count'),
                apps)

            return ApiResponse(
                response.status_code,
                data,
                raw=response)

        return ApiResponse(
            response.status_code,
            raw=response)

    # field methods

    def GetFieldById(self, fieldId: int):
        """
        Get a field by it's id.

        Parameters:
            fieldId (`int`): The unique id of the field being requested.

        Returns:
            An ApiResponse (`Models.ApiResponse`) containing the results of the request.
        """

        endpoint = GetFieldByIdEndpoint(self.baseUrl, fieldId)

        response = requests.request(
            'GET',
            endpoint,
            headers=self.headers)

        if response.status_code == 401:
            
            return ApiResponse(
                response.status_code,
                message='Unauthorized request',
                raw=response)

        if response.status_code == 403:
            
            return ApiResponse(
                response.status_code,
                message='Client does not have read access to the field',
                raw=response)

        if response.status_code == 404:
            
            return ApiResponse(
                response.status_code,
                message='Field could not be found',
                raw=response)

        if response.status_code == 200:

            jsonResponse = dict(response.json())

            values = jsonResponse.get('values')

            if values != None:

                listValues = []

                for value in values:
                    
                    value = dict(value)

                    value = ListValue(
                        value.get('id'),
                        value.get('name'),
                        value.get('sortOrder'),
                        value.get('numericValue'),
                        value.get('color'))

                    listValues.append(value)

            field = Field(
                jsonResponse.get('id'),
                jsonResponse.get('appId'),
                jsonResponse.get('name'),
                jsonResponse.get('type'),
                jsonResponse.get('status'),
                jsonResponse.get('isRequired'),
                jsonResponse.get('isUnique'),
                jsonResponse.get("listId"),
                listValues,
                jsonResponse.get('multiplicity'),
                jsonResponse.get('outputType'))

            data = GetFieldByIdResponse(field)

            return ApiResponse(
                response.status_code,
                data,
                raw=response)

        return ApiResponse(
            response.status_code,
            raw=response)

    def GetFieldsByIds(self, fieldIds: list):
        """
        Get a set of fields by their ids.

        Parameters:
            fieldIds (`list[int]`): A list of unique ids for the fields being requested.

        Returns:
            An ApiResponse (`Models.ApiResponse`) containing the results of the request.
        """

        endpoint = GetFieldsByIdsEndpoint(self.baseUrl)

        self.headers['Content-Type'] = 'application/json'

        # make sure fieldIds can be serialized to json string
        if not isinstance(fieldIds, (list, tuple)):
            return ApiResponse(
                400,
                message='Field ids should be of type list or tuple')

        fieldIds = json.dumps(fieldIds)

        response = requests.request(
            'POST',
            endpoint,
            headers=self.headers,
            data=fieldIds)

        if response.status_code == 401:
            
            return ApiResponse(
                response.status_code,
                message='Unauthorized request',
                raw=response)

        if response.status_code == 403:
            
            return ApiResponse(
                response.status_code,
                message='Client does not have read access to the field(s)',
                raw=response)

        if response.status_code == 404:
            
            return ApiResponse(
                response.status_code,
                message='Field(s) could not be found',
                raw=response)

        if response.status_code == 200:

            responseJson = dict(response.json())

            fields = []

            for item in responseJson.get('items'):
                
                item = dict(item)
                
                field = Field(
                    item.get('id'),
                    item.get('appId'),
                    item.get('name'),
                    item.get('type'),
                    item.get('status'),
                    item.get('isRequired'),
                    item.get('isUnique'))

                fields.append(field)

            data = GetFieldsByIdsResponse(
                responseJson.get('count'),
                fields)

            return ApiResponse(
                response.status_code,
                data,
                raw=response)

        return ApiResponse(
            response.status_code,
            raw=response)

    def GetFieldsByAppId(self, appId: int, pagingRequest=PagingRequest(1, 50)):
        """
        Get all fields for an app.

        Parameters:
            appId (`int`): The unique id of the app whose fields are being requested.
            pagingRequest (`Models.PagingRequest`): Used to set the page number and page size of the request. By default the these will be 1 and 50 respectively.

        Returns:
            An ApiResponse (`Models.ApiResponse`) containing the results of the request.
        """

        endpoint = GetFieldsByAppIdEndpoint(self.baseUrl, appId)

        params = pagingRequest.__dict__

        response = requests.request(
            'GET',
            endpoint,
            headers=self.headers,
            params=params)

        if response.status_code == 400:
            
            return ApiResponse(
                response.status_code,
                message='Invalid paging information',
                raw=response)

        if response.status_code == 401:
            
            return ApiResponse(
                response.status_code,
                message='Unauthorized request',
                raw=response)

        if response.status_code == 200:

            responseJson = dict(response.json())

            fields = []

            for item in responseJson.get('items'):

                item = dict(item)

                field = Field(
                    item.get('id'),
                    item.get('appId'),
                    item.get('name'),
                    item.get('type'),
                    item.get('status'),
                    item.get('isRequired'),
                    item.get('isUnique'))

                fields.append(field)

            data = GetFieldsByAppIdResponse(
                responseJson.get('pageNumber'),
                responseJson.get('pageSize'),
                responseJson.get('totalPages'),
                responseJson.get('totalRecords'),
                fields)

            return ApiResponse(
                response.status_code,
                data,
                raw=response)

        return ApiResponse(
            response.status_code,
            raw=response)

    # file methods

    def GetFileInfoById(self, recordId: int, fieldId: int, fileId: int):
        """
        Get the metadata information for a file.

        Parameters:
            recordId (`int`): The unique id of the record that contains the file you want to retrieve.
            fieldId (`int`): The unique id of the field that contains the file you want to retrieve.
            fileId (`int`): The unique id of the file whose metadata you want to retrieve.

        Returns:
            An ApiResponse (`Models.ApiResponse`) containing the results of the request.
        """

        endpoint = GetFileInfoByIdEndpoint(
            self.baseUrl,
            recordId,
            fieldId,
            fileId)

        response = requests.request(
            'GET',
            endpoint,
            headers=self.headers,)

        if response.status_code == 400:
            
            return ApiResponse(
                response.status_code,
                message='Request is invalid based on underlying data',
                raw=response)

        if response.status_code == 401:
            
            return ApiResponse(
                response.status_code,
                message='Unauthorized request',
                raw=response)

        if response.status_code == 403:
            
            return ApiResponse(
                response.status_code,
                message='Client does not have read access to the file',
                raw=response)

        if response.status_code == 404:
            
            return ApiResponse(
                response.status_code,
                message='File could not be found',
                raw=response)

        if response.status_code == 200:

            jsonResponse = dict(response.json())

            createdDate = parseDate(jsonResponse.get('createdDate'))
            modifiedDate = parseDate(jsonResponse.get('modifiedDate'))

            fileInfo = FileInfo(
                jsonResponse.get('type'),
                jsonResponse.get('contentType'),
                jsonResponse.get('name'),
                createdDate,
                modifiedDate,
                jsonResponse.get('owner'),
                jsonResponse.get('fileHref'))

            data = GetFileInfoByIdResponse(fileInfo)

            return ApiResponse(
                response.status_code,
                data,
                raw=response)

        return ApiResponse(
            response.status_code,
            raw=response)

    def DeleteFileById(self, recordId: int, fieldId: int, fileId: int):
        """
        Delete a file by its id.

        Parameters:
            recordId (`int`): The unique id of the record that contains the file you want to delete.
            fieldId (`int`): The unique id of the field that contains the file you want to delete.
            fileId (`int`): The unique id of the file you want to delete.
            
        Returns:
            An ApiResponse (`Models.ApiResponse`) containing the results of the request.
        """

        endpoint = DeleteFileByIdEndpoint(
            self.baseUrl,
            recordId,
            fieldId,
            fileId)

        response = requests.request(
            'DELETE',
            endpoint,
            headers=self.headers,)

        if response.status_code == 400:
            
            return ApiResponse(
                response.status_code,
                message='Request is invalid based on underlying data',
                raw=response)

        if response.status_code == 401:
            
            return ApiResponse(
                response.status_code,
                message='Unauthorized request',
                raw=response)

        if response.status_code == 403 or response.status_code == 404:

            jsonResponse = dict(response.json())

            return ApiResponse(
                response.status_code,
                message=jsonResponse.get('message'),
                raw=response)

        if response.status_code == 500:
            
            return ApiResponse(
                response.status_code,
                message='File could not be deleted due to internal error',
                raw=response)

        if response.status_code == 204:
            
            return ApiResponse(
                response.status_code,
                message='File deleted successfully',
                raw=response)

        return ApiResponse(
            response.status_code,
            raw=response)

    def GetFileById(self, recordId: int, fieldId: int, fileId: int):
        """
        Get a file by its id.

        Parameters:
            recordId (`int`): The unique id of the record that contains the file you want to retrieve.
            fieldId (`int`): The unique id of the field that contains the file you want to retrieve.
            fileId (`int`): The unique id of the file you want to retrieve.
            
        Returns:
            An ApiResponse (`Models.ApiResponse`) containing the results of the request.
        """

        endpoint = GetFileByIdEndpoint(
            self.baseUrl,
            recordId,
            fieldId,
            fileId)

        response = requests.request(
            'GET',
            endpoint,
            headers=self.headers,)

        if response.status_code == 400:
            
            return ApiResponse(
                response.status_code,
                message='Request is invalid based on underlying data',
                raw=response)

        if response.status_code == 401:
            
            return ApiResponse(
                response.status_code,
                message='Unauthorized request',
                raw=response)

        if response.status_code == 403 or response.status_code == 404:
            
            jsonResponse = dict(response.json())

            return ApiResponse(
                response.status_code,
                message=jsonResponse.get('message'),
                raw=response)

        if response.status_code == 200:

            headers = dict(response.headers)
            
            fileName = headers.get('Content-Disposition')
            result = re.search('filename=.*;', fileName).group()
            
            # TODO: implement attempting to build file name using content-type header
            if result:
                fileName = re.sub('filename=|\'|;', '', result)
            else:
                fileName = 'OnspringFile'

            file = File(
                fileName,
                headers.get('Content-Type'),
                headers.get('Content-Length'),
                response.content)

            data = GetFileByIdResponse(file)

            return ApiResponse(
                response.status_code,
                data,
                raw=response)
        

        return ApiResponse(
            response.status_code,
            raw=response)

    def SaveFile(self,  saveFileRequest: SaveFileRequest):
        """
        Delete a file by its id.

        Parameters:
            saveFileRequest (`Models.SaveFileRequest`): An object representing all the necessary information to make a successful request.
            
        Returns:
            An ApiResponse (`Models.ApiResponse`) containing the results of the request.
        """

        endpoint = SaveFileEndpoint(self.baseUrl)

        files = [
            (
                'File',
                (saveFileRequest.fileName, open(saveFileRequest.filePath,'rb'),saveFileRequest.contentType)
            )
        ]

        saveFileRequest = saveFileRequest.__dict__
        del saveFileRequest['fileName']
        del saveFileRequest['filePath']
        del saveFileRequest['contentType']

        requestData = saveFileRequest

        response = requests.request(
            'POST', 
            endpoint, 
            headers=self.headers,
            data=requestData,
            files=files)

        if response.status_code == 400:
            
            return ApiResponse(
                response.status_code,
                message='Request is invalid based on underlying data',
                raw=response)

        if response.status_code == 401:
            
            return ApiResponse(
                response.status_code,
                message='Unauthorized request',
                raw=response)

        if response.status_code in [403,404,500]:

            jsonResponse = dict(response.json())

            return ApiResponse(
                response.status_code,
                message=jsonResponse.get('message'),
                raw=response)

        if response.status_code == 201:

            responseJson = dict(response.json())

            data = SaveFileResponse(responseJson.get('id'))

            return ApiResponse(
                    response.status_code,
                    data,
                    raw=response)

        return ApiResponse(
            response.status_code,
            raw=response)

    # list methods

    def AddOrUpdateListItem(self, listItemRequest: ListItemRequest):
        """
        Add or update a list value by its id.

        Parameters:
            listItemRequest (`Models.ListItemRequest`): An object representing all the necessary information to make a successful request.
            
        Returns:
            An ApiResponse (`Models.ApiResponse`) containing the results of the request.
        """

        endpoint = AddOrUpdateListItemEndpoint(self.baseUrl, listItemRequest.listId)

        self.headers['Content-Type'] = 'application/json'

        del listItemRequest.__dict__['listId']

        requestData = json.dumps(listItemRequest.__dict__)

        response = requests.request(
            'PUT', 
            endpoint, 
            headers=self.headers,
            data=requestData)

        if response.status_code == 401:
            
            return ApiResponse(
                response.status_code,
                message='Unauthorized request',
                raw=response)

        if response.status_code in [403,404]:

            jsonResponse = dict(response.json())

            return ApiResponse(
                response.status_code,
                message=jsonResponse.get('message'),
                raw=response)

        if response.status_code == 201:
            
            responseJson = dict(response.json())

            data = AddOrUpdateListItemResponse(responseJson.get('id'))

            return ApiResponse(
                    response.status_code,
                    data,
                    message='New list value successfully added',
                    raw=response)

        if response.status_code == 200:
            
            responseJson = dict(response.json())

            data = AddOrUpdateListItemResponse(responseJson.get('id'))

            return ApiResponse(
                    response.status_code,
                    data,
                    message='Existing list value successfully updated',
                    raw=response)
        
        return ApiResponse(
            response.status_code,
            raw=response)

    def DeleteListItem(self, listId: int, itemId: str):
        """
        Delete a list value by its id and it's parent list id.

        Parameters:
            listId (`int`): The unique id of the list values parent list.
            itemId (`int`): The unique id of the list value to be deleted.
            
        Returns:
            An ApiResponse (`Models.ApiResponse`) containing the results of the request.
        """

        endpoint = DeleteListItemEndpoint(self.baseUrl, listId, itemId)

        response = requests.request(
            'DELETE',
            endpoint,
            headers=self.headers)

        if response.status_code == 401:
            
            return ApiResponse(
                response.status_code,
                message='Unauthorized request',
                raw=response)

        if response.status_code == 403:

            jsonResponse = dict(response.json())

            return ApiResponse(
                response.status_code,
                message=jsonResponse.get('message'),
                raw=response)

        if response.status_code == 404:
            
            return ApiResponse(
                response.status_code,
                message='List/item could not be found',
                raw=response)

        if response.status_code == 204:
            
            return ApiResponse(
                response.status_code,
                message='Item deleted successfully',
                raw=response)

        return ApiResponse(
            response.status_code,
            raw=response)

    # record methods

    def GetRecordsByAppId(self, getRecordsByAppRequest: GetRecordsByAppRequest):
        """
        Get all the records for an app by its id.

        Parameters:
            getRecordsByAppRequest (`Models.GetRecordsByAppRequest`): The unique id of the list values parent list.
            
        Returns:
            An ApiResponse (`Models.ApiResponse`) containing the results of the request.
        """

        endpoint = GetRecordsByAppIdEndpoint(self.baseUrl, getRecordsByAppRequest.appId)

        params = getRecordsByAppRequest.__dict__
        del params['appId']
        params['fieldIds'] = ",".join([str(i) for i in params['fieldIds']])

        response = requests.request(
            'GET',
            endpoint,
            headers=self.headers,
            params=params)

        if response.status_code == 400:
            
            return ApiResponse(
                response.status_code,
                message='Invalid paging information/size of the data requested was too large.',
                raw=response)

        if response.status_code == 401:
            
            return ApiResponse(
                response.status_code,
                message='Unauthorized request',
                raw=response)

        if response.status_code == 403:

            jsonResponse = dict(response.json())

            return ApiResponse(
                response.status_code,
                message=jsonResponse.get('message'),
                raw=response)

        if response.status_code == 200:

            jsonResponse = dict(response.json())

            records = []

            for item in jsonResponse.get('items'):
                
                item = dict(item)

                fields = []

                record = Record(
                    item.get('appId'),
                    fields,
                    item.get('recordId'))

                for field in item.get('fieldData'):
                    
                    field = dict(field)

                    field = RecordFieldValue(
                        field.get('fieldId'),
                        field.get('value'),
                        field.get('type'))
                    
                    fields.append(field)
            
                record.fields = fields

                records.append(record)

            data = GetRecordsResponse(
                jsonResponse.get('pageNumber'),
                jsonResponse.get('pageSize'),
                jsonResponse.get('totalPages'),
                jsonResponse.get('totalRecords'),
                records)

            return ApiResponse(
                response.status_code,
                data,
                raw=response)

        return ApiResponse(
            response.status_code,
            raw=response)
    
    def GetRecordById(self, getRecordByIdRequest: GetRecordByIdRequest):
        """
        Get a record by its id.

        Parameters:
            getRecordByIdRequest (`Models.GetRecordByIdRequest`): An object that contains all the necessary information for making a successful request.
            
        Returns:
            An ApiResponse (`Models.ApiResponse`) containing the results of the request.
        """

        endpoint = GetRecordByIdEndpoint(self.baseUrl, getRecordByIdRequest.appId, getRecordByIdRequest.recordId)

        params = getRecordByIdRequest.__dict__
        del params['appId']
        del params['recordId']
        params['fieldIds'] = ",".join([str(i) for i in params['fieldIds']])

        response = requests.request(
            'GET', 
            endpoint,
            headers=self.headers, 
            params=params)

        if response.status_code == 401:
            
            return ApiResponse(
                response.status_code,
                message='Unauthorized request',
                raw=response)

        if response.status_code == 403:

            jsonResponse = dict(response.json())

            return ApiResponse(
                response.status_code,
                message=jsonResponse.get('message'),
                raw=response)

        if response.status_code == 404:
            
            return ApiResponse(
                response.status_code,
                message='Record could not be found',
                raw=response)

        if response.status_code == 200:

            jsonResponse = dict(response.json())

            fields = []

            for field in jsonResponse.get('fieldData'):
                    
                field = dict(field)

                field = RecordFieldValue(
                    field.get('fieldId'),
                    field.get('value'),
                    field.get('type'))
                    
                fields.append(field)

            data = Record(
                jsonResponse.get('appId'),
                fields,
                jsonResponse.get('recordId'))

            return ApiResponse(
                response.status_code,
                data,
                raw=response)

        return ApiResponse(
            response.status_code,
            raw=response)

    def DeleteRecordById(self, appId: int, recordId: int):
        """
        Delete a record by its id.

        Parameters:
            appId (`int`): The id value of the app where the record you want to delete resides.
            recordId (`int`): The id value of the record you want to delete.
            
        Returns:
            An ApiResponse (`Models.ApiResponse`) containing the results of the request.
        """

        endpoint = DeleteRecordByIdEndpoint(self.baseUrl, appId, recordId)

        response = requests.request(
            'DELETE',
            endpoint,
            headers=self.headers)

        if response.status_code == 401:
            
            return ApiResponse(
                response.status_code,
                message='Unauthorized request',
                raw=response)

        if response.status_code == 403:

            jsonResponse = dict(response.json())

            return ApiResponse(
                response.status_code,
                message=jsonResponse.get('message'),
                raw=response)

        if response.status_code == 404:
            
            return ApiResponse(
                response.status_code,
                message='Record could not be found',
                raw=response)

        if response.status_code == 204:
            
            return ApiResponse(
                response.status_code,
                message='Record deleted successfully',
                raw=response)

        return ApiResponse(
            response.status_code,
            raw=response)
        
    def GetRecordsByIds(self, getBatchRecordsRequest: GetBatchRecordsRequest):
        """
        Get records by their id.

        Parameters:
            GetBatchRecordsRequest (`Models.GetBatchRecordsRequest`): An object that contains all the necessary information for making a successful request.
            
        Returns:
            An ApiResponse (`Models.ApiResponse`) containing the results of the request.
        """

        endpoint = GetRecordsByIdsEndpoint(self.baseUrl)

        self.headers['Content-Type'] = 'application/json'

        requestData = json.dumps(getBatchRecordsRequest.__dict__)

        response = requests.request(
            'POST',
            endpoint,
            headers=self.headers,
            data=requestData)

        if response.status_code == 400:
            
            return ApiResponse(
                response.status_code,
                message='Batch request is invalid/size of the data requested was too large.',
                raw=response)

        if response.status_code == 401:
            
            return ApiResponse(
                response.status_code,
                message='Unauthorized request',
                raw=response)

        if response.status_code == 403:

            jsonResponse = dict(response.json())

            return ApiResponse(
                response.status_code,
                message=jsonResponse.get('message'),
                raw=response)

        if response.status_code == 200:

            jsonResponse = dict(response.json())

            records = []

            for item in jsonResponse.get('items'):
                
                item = dict(item)

                fields = []

                record = Record(
                    item.get('appId'),
                    fields,
                    item.get('recordId'),)

                for field in item.get('fieldData'):
                    
                    field = dict(field)

                    field = RecordFieldValue(
                        field.get('fieldId'),
                        field.get('value'),
                        field.get('type'))
                    
                    fields.append(field)
            
                record.fields = fields

                records.append(record)

            data = GetBatchRecordsResponse(
                jsonResponse.get('count'),
                records)

            return ApiResponse(
                response.status_code,
                data,
                raw=response)

        return ApiResponse(
            response.status_code,
            raw=response)

    def QueryRecords(self, queryRecordsRequest: QueryRecordsRequest):
        """
        Get records based on a criteria.

        Parameters:
            queryRecordsRequest (`Models.QueryRecordsRequest`): An object that contains all the necessary information for making a successful request.
            
        Returns:
            An ApiResponse (`Models.ApiResponse`) containing the results of the request.
        """

        endpoint = QueryRecordsEndpoint(self.baseUrl)
        
        self.headers['Content-Type'] = 'application/json'

        requestData = queryRecordsRequest.__dict__

        params = requestData.get('pagingRequest').__dict__

        del requestData['pagingRequest']

        requestData = json.dumps(requestData)

        response = requests.request(
            'POST',
            endpoint,
            headers=self.headers,
            data=requestData,
            params=params)

        if response.status_code == 400:
            
            return ApiResponse(
                response.status_code,
                message='Query request is invalid/size of the data requested was too large.',
                raw=response)

        if response.status_code == 401:
            
            return ApiResponse(
                response.status_code,
                message='Unauthorized request',
                raw=response)

        if response.status_code == 403:

            jsonResponse = dict(response.json())

            return ApiResponse(
                response.status_code,
                message=jsonResponse.get('message'),
                raw=response)

        if response.status_code == 200:

            jsonResponse = dict(response.json())

            records = []

            for item in jsonResponse.get('items'):
                
                item = dict(item)

                fields = []

                record = Record(
                    item.get('appId'),
                    fields,
                    item.get('recordId'))

                for field in item.get('fieldData'):
                    
                    field = dict(field)

                    field = RecordFieldValue(
                        field.get('fieldId'),
                        field.get('value'),
                        field.get('type'))
                    
                    fields.append(field)
            
                record.fields = fields

                records.append(record)

            data = GetRecordsResponse(
                jsonResponse.get('pageNumber'),
                jsonResponse.get('pageSize'),
                jsonResponse.get('totalPages'),
                jsonResponse.get('totalRecords'),
                records)

            return ApiResponse(
                response.status_code,
                data,
                raw=response)

        return ApiResponse(
            response.status_code,
            raw=response)

    def AddOrUpdateRecord(self, record: Record):
        """
        Add a new record or update a record by its id. Not including an id adds a new record.

        Parameters:
            record (`Models.Record`): An object that contains all the information for making a successful request.
            
        Returns:
            An ApiResponse (`Models.ApiResponse`) containing the results of the request.
        """

        endpoint = AddOrUpdateRecordEndpoint(self.baseUrl)

        self.headers['Content-Type'] = 'application/json'

        fieldsDict = {}

        for field in record.fields:
            fieldsDict[field.fieldId] = field.value

        record.fields = fieldsDict

        requestData = json.dumps(record.__dict__, default=str)

        response = requests.request(
            'PUT',
            endpoint,
            headers=self.headers,
            data=requestData)

        if response.status_code == 400:
            
            return ApiResponse(
                response.status_code,
                message='Request is data is invalid',
                raw=response)

        if response.status_code == 401:
            
            return ApiResponse(
                response.status_code,
                message='Unauthorized request',
                raw=response)

        if response.status_code in [403, 404]:

            jsonResponse = dict(response.json())

            return ApiResponse(
                response.status_code,
                message=jsonResponse.get('message'),
                raw=response)

        if response.status_code in [200, 201]:

            jsonResponse = dict(response.json())

            data = AddOrUpdateRecordResponse(
                jsonResponse.get('id'),
                jsonResponse.get('warnings'))

            if response.status_code == 200:
                message = 'Record updated successfully'
            else:
                message = 'Record created successfully'

            return ApiResponse(
                response.status_code,
                data,
                message=message,
                raw=response)

        return ApiResponse(
                response.status_code,
                raw=response)

    def DeleteRecordsByIds(self, deleteBatchRecordsRequest: DeleteBatchRecordsRequest):
        """
        Delete records by their ids.

        Parameters:
            deleteBatchRecordsRequest (`Models.DeleteBatchRecordsRequest`): An object that contains all the necessary information for making a successful request.
            
        Returns:
            An ApiResponse (`Models.ApiResponse`) containing the results of the request.
        """

        endpoint = DeleteRecordsByIds(self.baseUrl)

        self.headers['Content-Type'] = 'application/json'

        requestData = json.dumps(deleteBatchRecordsRequest.__dict__)

        response = requests.request(
            'POST',
            endpoint,
            headers=self.headers,
            data=requestData)

        if response.status_code == 400:
            
            return ApiResponse(
                response.status_code,
                message='Invalid request provided',
                raw=response)

        if response.status_code == 401:
            
            return ApiResponse(
                response.status_code,
                message='Unauthorized request',
                raw=response)

        if response.status_code == 403:

            jsonResponse = dict(response.json())

            return ApiResponse(
                response.status_code,
                message=jsonResponse.get('message'),
                raw=response)

        if response.status_code == 404:
            
            return ApiResponse(
                response.status_code,
                message='Records could not be found',
                raw=response)

        if response.status_code == 204:
            
            return ApiResponse(
                response.status_code,
                message='Record(s) deleted successfully',
                raw=response)

        return ApiResponse(
            response.status_code,
            raw=response)

    # report methods

    def GetReportById(self, getReportByIdRequest: GetReportByIdRequest):
        """
        Get a report by its id.

        Parameters:
            getReportByIdRequest (`Models.GetReportByIdRequest`): An object that contains all the necessary information for making a successful request.
            
        Returns:
            An ApiResponse (`Models.ApiResponse`) containing the results of the request.
        """

        endpoint = GetReportByIdEndpoint(self.baseUrl, getReportByIdRequest.reportId)

        params = getReportByIdRequest.__dict__
        del params['reportId']

        response = requests.request(
            'GET',
            endpoint,
            headers=self.headers,
            params=params)
        
        if response.status_code == 400:
            
            return ApiResponse(
                response.status_code,
                message='Invalid request based on underlying data',
                raw=response)

        if response.status_code == 401:
            
            return ApiResponse(
                response.status_code,
                message='Unauthorized request',
                raw=response)

        if response.status_code == 403:

            jsonResponse = dict(response.json())

            return ApiResponse(
                response.status_code,
                message=jsonResponse.get('message'),
                raw=response)

        if response.status_code == 404:
            
            return ApiResponse(
                response.status_code,
                message='Report could not be found',
                raw=response)

        if response.status_code == 200:

            jsonResponse = dict(response.json())
            
            rows = []

            for row in jsonResponse.get('rows'):
                
                row = dict(row)
                
                row = Row(
                    row.get('recordId'),
                    row.get('cells'))
                
                rows.append(row)
            
            data = GetReportByIdResponse(
                jsonResponse.get('columns'),
                rows)

            return ApiResponse(
                response.status_code,
                data=data,
                raw=response)

        return ApiResponse(
            response.status_code,
            raw=response)

    def GetReportsByAppId(self, appId: int, pagingRequest: PagingRequest=PagingRequest(1,50)):
        """
        Get reports for an app by its id..

        Parameters:
            appId (`int`): The unique id of the app whose reports are being requested.
            pagingRequest (`Models.PagingRequest`): Used to set the page number and page size of the request. By default the these will be 1 and 50 respectively.

        Returns:
            An ApiResponse (`Models.ApiResponse`) containing the results of the request.
        """

        endpoint = GetReportsByAppIdEndpoint(self.baseUrl, appId)

        params = pagingRequest.__dict__

        response = requests.request(
            'GET',
            endpoint,
            headers=self.headers,
            params=params)

        if response.status_code == 400:
            return ApiResponse(
                response.status_code,
                message='Client does not have read access to the app.',
                raw=response)

        if response.status_code == 401:
            return ApiResponse(
                response.status_code,
                message='Unauthorized request',
                raw=response)

        if response.status_code == 403:

            jsonResponse = dict(response.json())

            return ApiResponse(
                response.status_code,
                message=jsonResponse.get('message'),
                raw=response)

        if response.status_code == 200:

            responseJson = dict(response.json())

            reports = []

            for item in responseJson.get('items'):

                item = dict(item)

                report = Report(
                    item.get('appId'),
                    item.get('id'),
                    item.get('name'),
                    item.get('description'))

                reports.append(report)

            data = GetReportsByAppIdResponse(
                responseJson.get('pageNumber'),
                responseJson.get('pageSize'),
                responseJson.get('totalPages'),
                responseJson.get('totalRecords'),
                reports)

            return ApiResponse(
                response.status_code,
                data,
                raw=response)

        return ApiResponse(
            response.status_code,
            raw=response)


import uuid

# connectivity endpoints

def GetPingEndpoint(baseUrl: str):
    """
    Returns the ping endpoint.

    Args:
        baseUrl (`str`): The base url for the api.

    Returns:
        The ping endpoint as a string.
    """
    
    return f'{baseUrl}/Ping'

# app endpoints

def GetAppsEndpoint(baseUrl: str):
    """
    Returns the get apps endpoint.

    Args:
        baseUrl (`str`): The base url for the api.

    Returns:
        The get apps endpoint as a string.
    """

    return f'{baseUrl}/Apps'

def GetAppByIdEndpoint(baseUrl: str, appId: int):
    """
    Returns the get app by id endpoint.

    Args:
        baseUrl (`str`): The base url for the api.
        appId (`int`): The id of the app being requested.

    Returns:
        The get app by id endpoint as a string.
    """

    return f'{baseUrl}/Apps/id/{appId}'

def GetAppsByIdsEndpoint(baseUrl: str):
    """
    Returns the get apps by ids endpoint.

    Args:
        baseUrl (`str`): The base url for the api.

    Returns:
        The get apps by ids endpoint as a string.
    """

    return f'{baseUrl}/Apps/batch-get'

# field endpoints

def GetFieldByIdEndpoint(baseUrl: str, fieldId: int):
    """
    Returns the get field by id endpoint.

    Args:
        baseUrl (`str`): The base url for the api.
        fieldId (`int`): The id of the field being requested.

    Returns:
        The get field by id endpoint as a string.
    """

    return f'{baseUrl}/Fields/id/{fieldId}'

def GetFieldsByIdsEndpoint(baseUrl: str):
    """
    Returns the get fields by ids endpoint.

    Args:
        baseUrl (`str`): The base url for the api.

    Returns:
        The get fields by ids endpoint as a string.
    """

    return f'{baseUrl}/Fields/batch-get'

def GetFieldsByAppIdEndpoint(baseUrl: str, appId: int):
    """
    Returns the get fields by app id endpoint.

    Args:
        baseUrl (`str`): The base url for the api.
        appId (`int`): The id of the app whose fields are being requested.

    Returns:
        The get fields by app id endpoint as a string.
    """

    return f'{baseUrl}/Fields/appId/{appId}'

# file endpoints

def GetFileInfoByIdEndpoint(baseUrl: str, recordId: int, fieldId: int, fileId: int):
    """
    Returns the get file info by its id endpoint.

    Args:
        baseUrl (`str`): The base url for the api.
        recordId (`int`): The id for the record where the file resides.
        fieldId (`int`): The id for the field in the record where the file is held.
        fileId ('int'): The id of the file.

    Returns:
        The get file info by its id endpoint as a string.
    """

    return f'{baseUrl}/Files/recordId/{recordId}/fieldId/{fieldId}/fileId/{fileId}'

def DeleteFileByIdEndpoint(baseUrl: str, recordId: int, fieldId: int, fileId: int):
    """
    Returns the delete file by its id endpoint.

    Args:
        baseUrl (`str`): The base url for the api.
        recordId (`int`): The id for the record where the file resides.
        fieldId (`int`): The id for the field in the record where the file is held.
        fileId ('int'): The id of the file.

    Returns:
        The delete file by its id endpoint as a string.
    """

    return f'{baseUrl}/Files/recordId/{recordId}/fieldId/{fieldId}/fileId/{fileId}'

def GetFileByIdEndpoint(baseUrl: str, recordId: int, fieldId: int, fileId: int):
    """
    Returns the get file by its id endpoint.

    Args:
        baseUrl (`str`): The base url for the api.
        recordId (`int`): The id for the record where the file resides.
        fieldId (`int`): The id for the field in the record where the file is held.
        fileId ('int'): The id of the file.

    Returns:
        The get file by its id endpoint as a string.
    """

    return f'{baseUrl}/Files/recordId/{recordId}/fieldId/{fieldId}/fileId/{fileId}/file'

def SaveFileEndpoint(baseUrl: str):
    """
    Returns the save file endpoint.

    Args:
        baseUrl (`str`): The base url for the api.

    Returns:
        The save file endpoint as a string.
    """

    return f'{baseUrl}/Files'

# list endpoints

def AddOrUpdateListItemEndpoint(baseUrl, listId: int):
    """
    Returns the add or update list item endpoint.

    Args:
        baseUrl (`str`): The base url for the api.
        listId (`int`): The id of the list for which the item belongs when being updated or being added to.

    Returns:
        The add or update list item endpoint as a string.
    """

    return f'{baseUrl}/Lists/id/{listId}/items'

def DeleteListItemEndpoint(baseUrl: str, listId: int, itemId: uuid):
    """
    Returns the delete list item endpoint.

    Args:
        baseUrl (`str`): The base url for the api.
        listId (`int`): The id of the list for which the list item belongs.
        itemId (`int`): The id list item.

    Returns:
        The delete list item endpoint as a string.
    """

    return f'{baseUrl}/Lists/id/{listId}/itemId/{itemId}'

# record endpoints

def GetRecordsByAppIdEndpoint(baseUrl, appId: int):
    """
    Returns the get records by app id endpoint.

    Args:
        baseUrl (`str`): The base url for the api.
        appId (`int`): The id of the app where the records reside.
    
    Returns:
        The get records by app id endpoint as a string.
    """

    return f'{baseUrl}/Records/appId/{appId}'

def GetRecordByIdEndpoint(baseUrl, appId: int, recordId: int):
    """
    Returns the get record by id endpoint.

    Args:
        baseUrl (`str`): The base url for the api.
        appId (`int`): The id of the app where the record resides.
        recordId (`int`): The id of the record.
    
    Returns:
        The get record by id endpoint as a string.
    """

    return f'{baseUrl}/Records/appId/{appId}/recordId/{recordId}'

def DeleteRecordByIdEndpoint(baseUrl, appId: int, recordId: int):
    """
    Returns the delete record by id endpoint.

    Args:
        baseUrl (`str`): The base url for the api.
        appId (`int`): The id of the app where the record resides.
        recordId (`int`): The id of the record.
    
    Returns:
        The delete record by id endpoint as a string.
    """

    return f'{baseUrl}/Records/appId/{appId}/recordId/{recordId}'

def GetRecordsByIdsEndpoint(baseUrl):
    """
    Returns the get records by ids endpoint.

    Args:
        baseUrl (`str`): The base url for the api.
    
    Returns:
        The get records by ids endpoint as a string.
    """
    
    return f'{baseUrl}/Records/batch-get'

def QueryRecordsEndpoint(baseUrl):
    """
    Returns the query records endpoint.

    Args:
        baseUrl (`str`): The base url for the api.
    
    Returns:
        The query records endpoint as a string.
    """

    return f'{baseUrl}/Records/Query'

def AddOrUpdateRecordEndpoint(baseUrl):
    """
    Returns the add or update record endpoint.

    Args:
        baseUrl (`str`): The base url for the api.
    
    Returns:
        The add or update record endpoint as a string.
    """

    return f'{baseUrl}/Records'

def DeleteRecordsByIds(baseUrl):
    """
    Returns the delete records by ids endpoint.

    Args:
        baseUrl (`str`): The base url for the api.
    
    Returns:
        The delete records by ids endpoint as a string.
    """

    return f'{baseUrl}/Records/batch-delete'

# report endpoints

def GetReportByIdEndpoint(baseUrl, reportId: int):
    """
    Returns the get report by id endpoint.

    Args:
        baseUrl (`str`): The base url for the api.
        reportId (`int`): The id of the report.
    
    Returns:
        The get report by id endpoint as a string.
    """

    return f'{baseUrl}/Reports/id/{reportId}'

def GetReportsByAppIdEndpoint(baseUrl, appId: int):
    """
    Returns the get reports by app id endpoint.

    Args:
        baseUrl (`str`): The base url for the api.
        appId (`int`): The id of the app where the reports reside.
    
    Returns:
        The get reports by app id endpoint as a string.
    """

    return f'{baseUrl}/Reports/appId/{appId}'


import datetime
import uuid

from OnspringApiSdk.Enums import *
from decimal import Decimal
from datetime import datetime
from OnspringApiSdk.Helpers import parseDate

# generic

class ApiResponse:
    """
    An object to represent a response to a request made by an `OnspringClient`.

    Attributes:
        statusCode (`int`): The http status code of the response.
        data: If the request was successful will contain the response data deserialized to custom python objects.
        message (`str`): A message that may provide more detail about the requests success or failure.
        raw (`requests.Response`): Exposes the raw response object of the request if you'd like to handle it directly.
    """

    def __init__(self, statusCode=None, data=None, message=None, raw=None):
        self.statusCode = statusCode
        self.isSuccessful = int(statusCode) < 400
        self.data = data
        self.message = message
        self.raw = raw

class PagingRequest:
    """
    An object to represent the page number and page size of a paginated request made by an `OnspringClient`.

    Attributes:
        pageNumber (`int`): The page number that will be requested.
        pageSize (`int`): The size of the page that will be requested.
    """

    def __init__(self, pageNumber: int, pageSize: int):
        self.pageNumber = pageNumber
        self.pageSize = pageSize

#app specific

class App:
    """
    An object to represent an Onspring app.

    Attributes:
        href (`str`): The href for the Onspring app.
        id (`int`): The id of the Onspring app.
        name (`str`): The name of the Onspring app.
    """

    def __init__(self, href: str, id: int, name: str):
        self.href = href
        self.id = id
        self.name = name

class GetAppsResponse:
    """
    An object to represent a paginated response to a request made by an `OnspringClient` to request a collection of Onspring apps.

    Attributes:
        pageNumber (`int`): The page number returned.
        pageSize (`int`): The size of the page returned.
        totalPages (`int`): The total number of pages for the request.
        totalRecords (`int`): The total records for the request.
        apps (`list[Models.App]`): The apps requested.
    """

    def __init__(self, pageNumber: int, pageSize: int, totalPages:int , totalRecords: int, apps: list[App]):
        self.pageNumber = pageNumber
        self.pageSize = pageSize
        self.totalPages = totalPages
        self.totalRecords = totalRecords
        self.apps = apps

class GetAppByIdResponse:
    """
    An object to represent a response to a request made by an `OnspringClient` to request an Onspring app.

    Attributes:
        app (`Models.App`): The requested app.
    """

    def __init__(self, app: App):
        self.app = app

class GetAppsByIdsResponse:
    """
    An object to represent a response to a request made by an `OnspringClient` to request a batch of Onspring apps.

    Attributes:
        count (`int`): The number of apps requested.
        apps (`list[Models.App]`): The apps requested.
    """

    def __init__(self, count: int, apps: list[App]):
        self.count = count
        self.apps = apps

# field specific

class ListValue:
    """
    An object to represent an Onspring list value.

    Attributes:
        id (`int`): The id of the Onspring list value.
        name (`str`): The list value's name.
        sortOrder (`int`): The list value's sort order in respect to other values in the same list.
        numericalValue (`Decimal`): The numeric value assigned to the list value.
        color (`str`): The color assigned to the list value.
    """

    def __init__(self, id: int, name: str, sortOrder: int, numericValue: Decimal, color: str):
        self.id = id
        self.name = name
        self.sortOrder = sortOrder

        if numericValue != None:
            self.numericValue = Decimal(numericValue)
        else:
            self.numericValue = numericValue
        
        self.color = color

    def AsString(self):
        """
        Gets the list value as a comma separated string of it's properties and their values.

        Args:
            None

        Returns:
            A `str` representation of the list value object.
        """

        return f'Id: {self.id}, Name: {self.name}, Value: {self.numericValue}, Sort Order: {self.sortOrder}, Color: {self.color}'

class Field:
    """
    An object to represent an Onspring field.

    Attributes:
        id (`int`): The id of the Onspring field.
        appId ('int'): The id of the Onspring app where the field resides.
        name (`str`): The name of the field.
        type (`str`): The type of the field.
        status (`str`): The stuat of the field.
        isRequired (`bool`): Indicates whether the field is required in Onspring or not.
        isUnique (`bool`): Indicates whether the field requires unique values in Onspring or not.
        listId (`int`): The id of the fields list if applicable. Used for list fields and formula fields with list output type.
        values (`list[Models.ListValue]`): The values of the fields list if applicable. Used for list fields and formula fields with list output type.
        multiplicity (`str`): The multiplicity of the field if applicable. Used for list fields, reference fields, and formula fields with a list output type.
        outputType (`str`): The output type of the field if applicable. Used for formula fields.
    """

    def __init__(
        self, id: int, 
        appId: int, 
        name: str, 
        type: str, 
        status: str, 
        isRequired: bool, 
        isUnique: bool, 
        listId: int=None, 
        values: list[ListValue]=None, 
        multiplicity: str=None,
        outputType: str=None
        ):

        self.id = id
        self.appId = appId
        self.name = name
        self.type = type
        self.status = status
        self.isRequired = isRequired
        self.isUnique = isUnique
        self.listId = listId
        self.values = values
        self.outputType = outputType
        self.multiplicity = multiplicity

class GetFieldByIdResponse:
    """
    An object to represent a response to a request made by an `OnspringClient` to request an Onspring field.

    Attributes:
        field (`Models.Field`): The requested field.
    """

    def __init__(self, field: Field):
        self.field = field

class GetFieldsByIdsResponse:
    """
    An object to represent a response to a request made by an `OnspringClient` to request a batch of Onspring fields.

    Attributes:
        count (`int`): The number of fields requested.
        fields (`list[Models.Field]`): The fields requested.
    """

    def __init__(self, count: int, fields: list[Field]):
        self.count = count
        self.fields = fields

class GetFieldsByAppIdResponse:
    """
    An object to represent a paginated response to a request made by an `OnspringClient` to request a collection of Onspring fields.

    Attributes:
        pageNumber (`int`): The page number returned.
        pageSize (`int`): The size of the page returned.
        totalPages (`int`): The total number of pages for the request.
        totalRecords (`int`): The total records for the request.
        fields (`list[Models.Field]`): The fields requested.
    """

    def __init__(self, pageNumber: int, pageSize: int, totalPages: int, totalRecords: int, fields: list[Field]):
        self.pageNumber = pageNumber
        self.pageSize = pageSize
        self.totalPages = totalPages
        self.totalRecords = totalRecords
        self.fields = fields

# file specific

class File:
    """
    An object to represent a file stored in Onsprng.

    Attributes:
        name (`str`): The name of the file.
        contentType (`str`): The content type of the file.
        contentLength (`int`): The length of the file's content.
        content (`bytes`): The content of the file.
    """

    def __init__(self, name: str, contentType: str, contentLength: int, content: bytes):
        self.name = name
        self.contentType = contentType
        self.contentLength = contentLength
        self.content = content

class FileInfo:
    """
    An object to represent the metadata for a file stored in Onspring.

    Attributes:
        type (`str`): The type of file as stored in Onspring Attachment or Image.
        contentType (`str`): The content type of the file itself.
        name (`str`): The name of the file
        createdDate (`datetime`): The created date of the file.
        modifiedDate (`datetime`): The modified date of the file.
        owner (`str`): The owner of the file.
        fileHref (`str`): The href for the file.
    """

    def __init__(self, type: str, contentType: str, name: str, createdDate: datetime, modifiedDate: datetime, owner: str, fileHref: str):
        self.type = type
        self.contentType = contentType
        self.name = name
        self.createdDate = createdDate
        self.modifiedDate = modifiedDate
        self.owner = owner
        self.fileHref = fileHref

class GetFileInfoByIdResponse:
    """
    An object to represent a response to a request made by an `OnspringClient` to request a info for a file in Onspring.

    Attributes:
        fileInfo (`Models.FileInfo`): The file info requested.
    """

    def __init__(self, fileInfo: FileInfo):
        self.fileInfo = fileInfo

class GetFileByIdResponse:
    """
    An object to represent a response to a request made by an `OnspringClient` to request a file in Onspring.

    Attributes:
        file (`Models.File`): The file requested.
    """

    def __init__(self, file: File):
        self.file = file

class SaveFileRequest:
    """
    An object to represent all the necessary information needed to make a successful 'OnspringClient.SaveFile' request.

    Attributes:
        recordId (`int`): The unique id of the record where you want to save the file.
        fieldId (`int`): The unique id of the field where you want to save the file.
        fileName (`str`): The name of the file you want to save.
        filePath (`str`): The path to the file you want to save.
        contentType (`str`): The content type of the file you want to save.
        notes (`str`): An option note about the file.
        notes (`datetime`): An optional date noting when the file was modified.
    """
    def __init__(self, recordId: int, fieldId: int, fileName: str, filePath: str, contentType: str, notes: str=None, modifiedDate: datetime=None):
        self.recordId = recordId
        self.fieldId = fieldId
        self.notes = notes
        self.modifiedDate = modifiedDate
        self.fileName = fileName
        self.filePath = filePath
        self.contentType = contentType

class SaveFileResponse:
    """
    An object to represent a response to a request made by an `OnspringClient` to save a file in Onspring.

    Attributes:
        id (`int`): The id of the file saved in Onspring.
    """

    def __init__(self, id: int):
        self.id = id

# list specific

class ListItemRequest:
    """
    An object to represent the necessary information for adding or updating a list value. If no id is provided the list value will be added. If an id is provided then an attempt will be made to find that list value and update it.

    Attributes:
        listId (`int`): The id of the parent list that the list value belongs to.
        name (`str`): The name of the list value.
        id (`uuid`): The id of the list value.
        numericValue (`int`): The numeric value assigned to the list value.
        color (`str`): The color value assigned to the list value.
    """

    def __init__(self, listId: int, name: str, id: uuid=None, numericValue: int=None, color: str=None):
        self.listId = listId
        self.name = name
        self.id = id
        self.numericValue = numericValue
        self.color = color

class AddOrUpdateListItemResponse:
    """
    An object to represent a response to a request made by an `OnspringClient` to add or update a list value in Onspring.

    Attributes:
        id (`int`): The id of the list value updated or added in Onspring.
    """

    def __init__(self, id: uuid):
        self.id = id

# record specific

class RecordFieldValue:
    """
    An object to represent the value in a field in an Onspring record.

    Attributes:
        fieldId (`int`): The id of the field that the value is in.
        value (`str`): The value of the field.
        type (`str`): The type of value.
    """

    def __init__(self, fieldId: int, value: str, type: str=None):
        self.fieldId = fieldId
        self.value = value
        self.type = type

    def AsString(self):
        """
        If the `Models.RecordFieldValue` type is String will return the value property as a `str` otherwise will return `None`.

        Args:
            None

        Returns:
            The value of the `Models.RecordFieldValue` as a `str`.
        """
        
        if self.type != ResultValueType.String.name:
            return None

        return StringFieldValue(self.fieldId, self.value).value

    def AsInteger(self):
        """
        If the `Models.RecordFieldValue` type is Integer will return the value property as an `int` otherwise will return `None`.

        Args:
            None

        Returns:
            The value of the `Models.RecordFieldValue` as an `int`.
        """
        
        if self.type != ResultValueType.Integer.name:
            return None
        
        return IntegerFieldValue(self.fieldId, int(self.value)).value

    def AsDecimal(self):
        """
        If the `Models.RecordFieldValue` type is Decimal will return the value property as a `Decimal` otherwise will return `None`.

        Args:
            None

        Returns:
            The value of the `Models.RecordFieldValue` as a `Decimal`.
        """
        
        if self.type != ResultValueType.Decimal.name:
            return None
        
        return DecimalFieldValue(self.fieldId, Decimal(self.value)).value

    def AsDate(self):
        """
        If the `Models.RecordFieldValue` type is Date will return the value property as a `datetime` otherwise will return `None`.

        Args:
            None

        Returns:
            The value of the `Models.RecordFieldValue` as a `datetime`.
        """
        
        if self.type != ResultValueType.Date.name:
            return None

        date = parseDate(self.value)

        return DateFieldValue(self.fieldId, date).value

    def AsGuid(self):
        """
        If the `Models.RecordFieldValue` type is Guid will return the value property as an `UUID` otherwise will return `None`.

        Args:
            None

        Returns:
            The value of the `Models.RecordFieldValue` as an `UUID`.
        """

        if self.type != ResultValueType.Guid.name:
            return None
        
        return GuidFieldValue(self.fieldId, uuid.UUID(self.value)).value
        
    def AsTimeSpan(self):
        """
        If the `Models.RecordFieldValue` type is TimeSpan will return the value property as a `Model.TimeSpanData` otherwise will return `None`.

        Args:
            None

        Returns:
            The value of the `Models.RecordFieldValue` as a `Model.TimeSpanData`.
        """

        if self.type != ResultValueType.TimeSpan.name:
            return None
        
        value = dict(self.value)

        quantity = value.get('quantity')
        increment = value.get('increment')
        recurrence = value.get('recurrence')
        endByDate = value.get('endByDate')
        endAfterOccurrences = value.get('endAfterOccurrences')

        endByDate = parseDate(endByDate)

        data = TimeSpanData(
            quantity,
            increment,
            recurrence,
            endByDate,
            endAfterOccurrences)

        return TimeSpanValue(self.fieldId, data).value

    def AsStringList(self):
        """
        If the `Models.RecordFieldValue` type is StringList will return the value property as a `list[str]` otherwise will return `None`.

        Args:
            None

        Returns:
            The value of the `Models.RecordFieldValue` as a `list[str]`.
        """

        if self.type != ResultValueType.StringList.name:
            return None
        
        strings = [str(string) for string in self.value]

        return StringListValue(self.fieldId, strings).value

    def AsIntegerList(self):
        """
        If the `Models.RecordFieldValue` type is IntegerList will return the value property as a `list[int]` otherwise will return `None`.

        Args:
            None

        Returns:
            The value of the `Models.RecordFieldValue` as a `list[int]`.
        """
        
        if self.type != ResultValueType.IntegerList.name:
            return None

        integers = [int(integer) for integer in self.value]
            
        return IntegerListValue(self.fieldId, integers).value

    def AsGuidList(self):
        """
        If the `Models.RecordFieldValue` type is GuidList will return the value property as a `list[UUID]` otherwise will return `None`.

        Args:
            None

        Returns:
            The value of the `Models.RecordFieldValue` as a `list[UUID]`.
        """

        if self.type != ResultValueType.GuidList.name:
            return None

        guids = []

        for guid in self.value:
            guids.append(uuid.UUID(guid))
            
        return GuidListValue(self.fieldId, guids).value

    def AsAttachmentList(self):
        """
        If the `Models.RecordFieldValue` type is AttachmentList will return the value property as a `list[Models.Attachment]` otherwise will return `None`.

        Args:
            None

        Returns:
            The value of the `Models.RecordFieldValue` as a `list[Models.Attachment]`.
        """

        if self.type != ResultValueType.AttachmentList.name:
            return None

        attachments = []

        for attachment in self.value:

            attachment = dict(attachment)

            attachment = Attachment(
                attachment.get('fileId'),
                attachment.get('fileName'),
                attachment.get('notes'),
                attachment.get('storageLocation'))

            attachments.append(attachment)

        return AttachmentListValue(self.fieldId, attachments).value

    def AsScoringGroupList(self):
        """
        If the `Models.RecordFieldValue` type is ScoringGroupList will return the value property as a `list[Models.ScoringGroup]` otherwise will return `None`.

        Args:
            None

        Returns:
            The value of the `Models.RecordFieldValue` as a `list[Models.ScoringGroup]`.
        """

        if self.type != ResultValueType.ScoringGroupList.name:
            return

        scoringGroups = []

        for scoringGroup in self.value:

            scoringGroup = dict(scoringGroup)

            scoringGroup = ScoringGroup(
                uuid.UUID(scoringGroup.get('listValueId')),
                scoringGroup.get('name'),
                Decimal(scoringGroup.get('score')),
                Decimal(scoringGroup.get('maximumScore')))

            scoringGroups.append(scoringGroup)
        
        return ScoringGroupListValue(self.fieldId, scoringGroups).value

    def AsFileList(self):
        """
        If the `Models.RecordFieldValue` type is FileList will return the value property as a `list[int]` otherwise will return `None`.

        Args:
            None

        Returns:
            The value of the `Models.RecordFieldValue` as a `list[int]`.
        """
        
        if self.type != ResultValueType.FileList.name:
            return None

        files = [int(file) for file in self.value]
            
        return FileListValue(self.fieldId, files).value

    def getValue(self):
        """
        Will determine the appropriate way to return the fields value based on it's type.

        Args:
            None

        Returns:
            The value of the `Models.RecordFieldValue` as the appropriate object.
        """

        if self.type == ResultValueType.String.name:
            return self.AsString()

        elif self.type == ResultValueType.Integer.name:
            return self.AsInteger()

        elif self.type == ResultValueType.Decimal.name:
            return self.AsDecimal()

        elif self.type == ResultValueType.Date.name:
            return self.AsDate()

        elif self.type == ResultValueType.TimeSpan.name:
            return self.AsTimeSpan()

        elif self.type == ResultValueType.Guid.name:
            return self.AsGuid()

        elif self.type == ResultValueType.StringList.name:
            return self.AsStringList()

        elif self.type == ResultValueType.IntegerList.name:
            return self.AsIntegerList()

        elif self.type == ResultValueType.GuidList.name:
            return self.AsGuidList()

        elif self.type == ResultValueType.AttachmentList.name:
            return self.AsAttachmentList()

        elif self.type == ResultValueType.ScoringGroupList.name:
            return self.AsScoringGroupList()

        elif self.type == ResultValueType.FileList.name:
            return self.AsFileList()

        else:
            return None

    def GetResultValueString(self):
        """
        Will return the value property regardless of the field value's type as a string.

        Args:
            None

        Returns:
            The value of the `Models.RecordFieldValue` as a string.
        """
        if self.type == ResultValueType.String.name:
            return self.AsString()

        elif self.type == ResultValueType.Integer.name:
            return self.AsInteger()

        elif self.type == ResultValueType.Decimal.name:
            return self.AsDecimal()

        elif self.type == ResultValueType.Date.name:
            return self.AsDate()

        elif self.type ==  ResultValueType.TimeSpan.name:
            data = self.AsTimeSpan()
            return f'Quantity: {data.quantity}, Increment: {data.increment}, Recurrence: {data.recurrence}, EndByDate: {data.endByDate}, EndAfterOccurrences: {data.endAfterOccurrences}'
            
        elif self.type == ResultValueType.Guid.name:
            return self.AsGuid()
            
        elif self.type ==  ResultValueType.StringList.name:
            data = self.AsStringList()
            return f'{",".join(data)}'
            
        elif self.type == ResultValueType.IntegerList.name:
            data = self.AsIntegerList()
            return f'{",".join([str(i) for i in data])}'
            
        elif self.type == ResultValueType.GuidList.name:
            data = self.AsGuidList()
            return f'{",".join([str(guid) for guid in data])}'
            
        elif self.type == ResultValueType.AttachmentList.name:
            data = self.AsAttachmentList()

            strings = []

            for attachment in data:
                string = f'FileId: {attachment.fileId}, FileName: {attachment.fileName}, Notes: {attachment.notes}, StorageLocation: {attachment.storageLocation}'
                strings.append(string)
            
            return f'{"; ".join(strings)}'
            
        elif self.type == ResultValueType.ScoringGroupList.name:
            data = self.AsScoringGroupList()

            strings = []

            for scoringGroup in data:
                string = f'ListValueId: {scoringGroup.listValueId}, Name: {scoringGroup.name}, Score: {scoringGroup.score}, Max Score: {scoringGroup.maximumScore}'
                strings.append(string)

            return f'{"; ".join(strings)}'

        elif self.type == ResultValueType.FileList.name:
            data = self.AsFileList()
            return f'{",".join([str(i) for i in data])}'

        else:
            return None

class Record:
    """
    An object to represent an Onspring record.

    Attributes:
        appId ('int'): The id of the Onspring app where the record resides.
        recordId (`int`): The id of the Onspring record.
        fields (`list[Models.RecordFieldValue]`): The record's field values.
    """
    
    def __init__(self, appId: int, fields: list[RecordFieldValue], recordId: int=None):
        self.appId = appId
        self.recordId = recordId
        self.fields = fields

class GetRecordsByAppRequest:
    """
    An object to represent all the necessary information for making a succcessful request to get a collection of Onspring records.

    Attributes:
        appId (`int`): The id for the Onspring app where the records reside.
        fieldIds (`list[int]`): The ids for the fields in the Onspring app that should be included for each record in the response.
        dataFormat (`str`): The format of the response data.
        pagingRequest (`Models.PagingRequest`): Used to set the page number and page size of the request. By default the these will be 1 and 50 respectively.
    """

    def __init__(self, appId: int, fieldIds: list[int]=[], dataFormat: str=DataFormat.Raw.name, pagingRequest: PagingRequest=PagingRequest(1,50)):
        self.appId = appId
        self.fieldIds = fieldIds
        self.dataFormat = dataFormat
        self.pageSize = pagingRequest.pageSize
        self.pageNumber = pagingRequest.pageNumber

class QueryRecordsRequest:
    """
    An object to represent all the necessary information for making a succcessful request to get a collection of Onspring records based on a specific criteria. For more information on constructing a proper filter please refer to the official Onspring API guide: https://shorturl.at/cnsFK.

    Attributes:
        appId (`int`): The id for the Onspring app where the records reside.
        filter (`str`): The criteria used to determine what records should be included in the response.
        fieldIds (`list[int]`): The ids for the fields in the Onspring app that should be included for each record in the response.
        dataFormat (`str`): The format of the response data.
        pagingRequest (`Models.PagingRequest`): Used to set the page number and page size of the request. By default the these will be 1 and 50 respectively.
    """

    def __init__(self, appId: int, filter: str, fieldIds: list[int]=[], dataFormat: str=DataFormat.Raw.name, pagingRequest: PagingRequest=PagingRequest(1,50)):
        self.appId = appId
        self.filter = filter
        self.fieldIds = fieldIds
        self.dataFormat = dataFormat
        self.pagingRequest = pagingRequest

class GetRecordsResponse:
    """
    An object to represent a paginated response to a request made by an `OnspringClient` to request a collection of Onspring records.

    Attributes:
        pageNumber (`int`): The page number returned.
        pageSize (`int`): The size of the page returned.
        totalPages (`int`): The total number of pages for the request.
        totalRecords (`int`): The total records for the request.
        records (`list[Models.Record]`): The records requested.
    """

    def __init__(self, pageNumber: int, pageSize: int, totalPages: int, totalRecords: int, records: list[Record]):
        self.pageNumber = pageNumber
        self.pageSize = pageSize
        self.totalPages = totalPages
        self.totalRecords = totalRecords
        self.records = records

class GetRecordByIdRequest:
    """
    An object to represent all the necessary information for making a succcessful request to get an Onspring record by its id.

    Attributes:
        appId (`int`): The id for the Onspring app where the record resides.
        recordId (`int`): The id for the record being requested.
        fieldIds (`list[int]`): The ids for the fields in the Onspring app that should be included for each record in the response.
        dataFormat (`str`): The format of the response data.
    """
    def __init__(self, appId: int, recordId: int, fieldIds: list[int]=[], dataFormat: str=DataFormat.Raw.name):
        self.appId = appId
        self.recordId = recordId
        self.fieldIds = fieldIds
        self.dataFormat = dataFormat

class GetBatchRecordsRequest:
    """
    An object to represent all the necessary information for making a succcessful request to get a batch of Onspring records by their ids.

    Attributes:
        appId (`int`): The id for the Onspring app where the records reside.
        recordId (`list[int]`): The ids for the records being requested.
        fieldIds (`list[int]`): The ids for the fields in the Onspring app that should be included for each record in the response.
        dataFormat (`str`): The format of the response data.
    """

    def __init__(self, appId: int, recordIds: list[int], fieldIds: list[int]=[], dataFormat: str=DataFormat.Raw.name):
        self.appId = appId
        self.recordIds = recordIds
        self.fieldIds = fieldIds
        self.dataFormat = dataFormat

class GetBatchRecordsResponse:
    """
    An object to represent a response to a request made by an `OnspringClient` to request a batch of Onspring records.

    Attributes:
        count (`int`): The number of records requested.
        records (`list[Models.Record]`): The records requested.
    """

    def __init__(self, count: int, records: list[Record]):
        self.count = count
        self.records = records

class AddOrUpdateRecordResponse:
    """
    An object to represent a response to a request made by an `OnspringClient` to add or update an Onspring record.
    
    Attributes:
        id (`int`): The id of the Onspring record that was added or updated.
        warnings ('list[str]'): A list of warnings.
    """

    def __init__(self, id: int, warnings: list[str]=[]):
        self.id = id
        self.warnings = warnings

class DeleteBatchRecordsRequest:
    """
    An object to represent all the necessary information for making a succcessful request to delete a batch of Onspring records by their ids.

    Attributes:
        appId (`int`): The id for the Onspring app where the records reside.
        recordId (`list[int]`): The ids for the records being deleted.
    """

    def __init__(self, appId: int, recordIds: list[int]):
        self.appId = appId
        self.recordIds = recordIds

# field types

class StringFieldValue(RecordFieldValue):
    """
    An object to represent an Onspring field value of the String type.

    Attributes:
        fieldId (`int`): The id of the field that the value is in.
        value (`str`): The value of the field.
        type (`str`): The type of value.
    """

    def __init__(self, fieldId: int, value):
        self.type = ResultValueType.String.name
        RecordFieldValue.__init__(self, fieldId, value, self.type)

class IntegerFieldValue(RecordFieldValue):
    """
    An object to represent an Onspring field value of the Integer type.

    Attributes:
        fieldId (`int`): The id of the field that the value is in.
        value (`int`): The value of the field.
        type (`str`): The type of value.
    """

    def __init__(self, fieldId: int, value: int):
        self.type = ResultValueType.Integer.name
        RecordFieldValue.__init__(self, fieldId, value, self.type)

class DecimalFieldValue(RecordFieldValue):
    """
    An object to represent an Onspring field value of the Decimal type.

    Attributes:
        fieldId (`int`): The id of the field that the value is in.
        value (`Decimal`): The value of the field.
        type (`str`): The type of value.
    """

    def __init__(self, fieldId: int, value: Decimal):
        self.type = ResultValueType.Decimal.name
        RecordFieldValue.__init__(self, fieldId, value, self.type)

class DateFieldValue(RecordFieldValue):
    """
    An object to represent an Onspring field value of the Date type.

    Attributes:
        fieldId (`int`): The id of the field that the value is in.
        value (`datetime`): The value of the field.
        type (`str`): The type of value.
    """

    def __init__(self, fieldId: int, value: datetime):
        self.type = ResultValueType.Date.name
        RecordFieldValue.__init__(self, fieldId, value, self.type)

class GuidFieldValue(RecordFieldValue):
    """
    An object to represent an Onspring field value of the Guid type.

    Attributes:
        fieldId (`int`): The id of the field that the value is in.
        value (`UUID`): The value of the field.
        type (`str`): The type of value.
    """

    def __init__(self, fieldId: int, value: uuid.UUID):
        self.type = ResultValueType.Guid.name
        RecordFieldValue.__init__(self, fieldId, value, self.type)

class TimeSpanData:
    """
    An object to represent the data that makes up an Onspring timespan field.

    Attributes:
        quantity (`Decimal`):
        increment (`Enums.Increment`):
        recurrence (`Enums.Recurrence`):
        endByDate (`datetime`):
        endAfterOccurrences (`int`):
    """

    def __init__(self, quantity: Decimal, increment: Increment, recurrence: Recurrence=None, endByDate: datetime=None, endAfterOccurrences: int=None):
        self.quantity = quantity
        self.increment = increment
        self.recurrence = recurrence
        self.endByDate = endByDate
        self.endAfterOccurrences = endAfterOccurrences

    def AsString(self):
        if self.endByDate != None:
            formattedDate = self.endByDate.strftime("%m/%d/%Y %I:%M %p")
            return f'Every {self.quantity} {self.increment} End By {formattedDate}'
        elif self.endAfterOccurrences != None:
            return f'Every {self.quantity} {self.increment} End After {self.endAfterOccurrences}'
        else:
            return f'{self.quantity} {self.increment}'

class TimeSpanValue(RecordFieldValue):
    """
    An object to represent an Onspring field value of the TimeSpan type.

    Attributes:
        fieldId (`int`): The id of the field that the value is in.
        value (`Models.TimeSpanData`): The value of the field.
        type (`str`): The type of value.
    """

    def __init__(self, fieldId: int, value: TimeSpanData):
        self.type = ResultValueType.TimeSpan.name
        RecordFieldValue.__init__(self, fieldId, value, self.type)

class StringListValue(RecordFieldValue):
    """
    An object to represent an Onspring field value of the StringList type.

    Attributes:
        fieldId (`int`): The id of the field that the value is in.
        value (`list[str]`): The value of the field.
        type (`str`): The type of value.
    """

    def __init__(self, fieldId: int, value: list[str]):
        self.type = ResultValueType.StringList.name
        RecordFieldValue.__init__(self, fieldId, value, self.type)

class IntegerListValue(RecordFieldValue):
    """
    An object to represent an Onspring field value of the IntegerList type.

    Attributes:
        fieldId (`int`): The id of the field that the value is in.
        value (`list[int]`): The value of the field.
        type (`str`): The type of value.
    """

    def __init__(self, fieldId: int, value: list[int]):
        self.type = ResultValueType.IntegerList.name
        RecordFieldValue.__init__(self, fieldId, value, self.type)

class GuidListValue(RecordFieldValue):
    """
    An object to represent an Onspring field value of the GuidList type.

    Attributes:
        fieldId (`int`): The id of the field that the value is in.
        value (`list[UUID]`): The value of the field.
        type (`str`): The type of value.
    """

    def __init__(self, fieldId: int, value: list[uuid.UUID]):
        self.type = ResultValueType.GuidList.name
        RecordFieldValue.__init__(self, fieldId, value, self.type)

class Attachment:
    """
    An object to represent an attachment in Onspring.

    Attributes:
        fileId (`int`): The id of the file in Onspring.
        fileName (`str`): The name of the file in Onspring.
        notes (`str`): The notes for the file in Onspring.
        storageLocation (`str`): The storage location of the file in Onspring.
    """

    def __init__(self, fileId: int, fileName: str, notes: str, storageLocation: str):
        self.fileId = fileId
        self.fileName = fileName
        self.notes = notes
        self.storageLocation = storageLocation

class AttachmentListValue(RecordFieldValue):
    """
    An object to represent an Onspring field value of the AttachmentList type.

    Attributes:
        fieldId (`int`): The id of the field that the value is in.
        value (`list[Models.Attachment]`): The value of the field.
        type (`str`): The type of value.
    """

    def __init__(self, fieldId: int, value: list[Attachment]):
        self.type = ResultValueType.AttachmentList.name
        RecordFieldValue.__init__(self, fieldId, value, self.type)

class FileListValue(RecordFieldValue):
    """
    An object to represent an Onspring field value of the FileList type.

    Attributes:
        fieldId (`int`): The id of the field that the value is in.
        value (`list[int]`): The value of the field.
        type (`str`): The type of value.
    """

    def __init__(self, fieldId: int, value: list[int]):
        self.type = ResultValueType.FileList.name
        RecordFieldValue.__init__(self, fieldId, value, self.type)

class ScoringGroup:
    """
    An object to represent an Onspring scoring group.

    Attributes:
        listValueId (`UUID`): The id of the list value.
        name (`str`): The name of the list value.
        score (`Decimal`): The score for the list value.
        maximumScore (`Decimal`): The maximum possible score for the group.
    """

    def __init__(self, listValueId: uuid.UUID, name: str, score: Decimal, maximumScore: Decimal):
        self.listValueId = listValueId
        self.name = name
        self.score = score
        self.maximumScore = maximumScore

class ScoringGroupListValue(RecordFieldValue):
    """
    An object to represent an Onspring field value of the ScoringGroupList type.

    Attributes:
        fieldId (`int`): The id of the field that the value is in.
        value (`list[Models.ScoringGroup]`): The value of the field.
        type (`str`): The type of value.
    """

    def __init__(self, fieldId: int, value: list[ScoringGroup]):
        self.type = ResultValueType.ScoringGroupList.name
        RecordFieldValue.__init__(self, fieldId, value, self.type)

# report specific

class GetReportByIdRequest:
    """
    An object to represent the necessary information to make a successful request to get a report by it's id.

    Attributes:
        reportId (`int`): The id of the report.
        apiDataFormat (`str`): The format of the data in the report.
        dataType (`str`): The data type for the report.
    """
    
    def __init__(self, reportId: int, apiDataFormat: str=DataFormat.Raw.name, dataType: str=ReportDataType.ReportData.name):
        self.reportId = reportId
        self.apiDataFormat = apiDataFormat
        self.dataType = dataType

class Row:
    """
    An object to represent a row of an Onspring report.
    
    Attributes:
        recordId (`int`): The id of the record who's data is held in the row.
        cells (`list[str]`): The record field values held in the row.
    """

    def __init__(self, recordId: int, cells: list[str]):
        self.recordId = recordId
        self.cells = cells

class GetReportByIdResponse:
    """
    An object to represent the response to a request made by an `OnspringClient` to get a report by its id.

    Attributes:
        columns (`list[int]`): Values indicating the columns in the report.
        rows (`list[Models.Row]`): A collection of rows representing the records in the report.
    """

    def __init__(self, columns: list[str], rows: list[Row]):
        self.columns = columns
        self.rows = rows

class Report:
    """
    An object to represent an Onspring report.

    Attributes:
        appId (`int`): The id of the app the report resides in.
        id (`int`): The id of the report.
        name (`str`): The name of the report.
        description (`str`): The description for the report.
    """

    def __init__(self, appId: int, id: int, name: str, description: str):
        self.appId = appId
        self.id = id
        self.name = name
        self.description = description

class GetReportsByAppIdResponse:
    """
    An object to represent the response to a request made by an `OnspringClient` to get a collection of reports by an app id.

    Attributes:
        pageNumber (`int`): The page number returned.
        pageSize (`int`): The size of the page returned.
        totalPages (`int`): The total number of pages for the request.
        totalRecords (`int`): The total records for the request.
        reports (`list[Models.Report]`): The requested reports for the current page.
    """

    def __init__(self, pageNumber: int, pageSize: int, totalPages: int, totalRecords: int, reports: list[Report]):
        self.pageNumber = pageNumber
        self.pageSize = pageSize
        self.totalPages = totalPages
        self.totalRecords = totalRecords
        self.reports = reports


from datetime import datetime
from OnspringApiSdk.Enums import *

def parseDate(date: str):

    if date==None:
        return None

    for format in ["%Y-%m-%dT%H:%M:%S.%fZ","%Y-%m-%dT%H:%M:%SZ"]:
            try:
                return datetime.strptime(date, format)
            except ValueError:
                pass